# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.core._util import CLIError
from azure.cli.core.commands.parameters import get_resources_in_subscription

from ._constants import (
    ACR_RESOURCE_PROVIDER,
    ACR_RESOURCE_TYPE,
    STORAGE_RESOURCE_TYPE
)
from ._factory import (
    get_arm_service_client,
    get_storage_service_client,
    get_acr_service_client,
    get_acr_api_version
)

def _arm_get_resource_by_name(resource_name, resource_type):
    '''Returns the ARM resource in the current subscription with resource_name.
    :param str resource_name: The name of resource
    :param str resource_type: The type of resource
    '''
    result = get_resources_in_subscription(resource_type)
    elements = [item for item in result if item.name.lower() == resource_name.lower()]

    if len(elements) == 0:
        raise CLIError(
            'No resource with type {} can be found with name: {}'.format(
                resource_type, resource_name))
    elif len(elements) == 1:
        return elements[0]
    else:
        raise CLIError(
            'More than one resources with type {} are found with name: {}'.format(
                resource_type, resource_name))

def get_resource_group_name_by_resource_id(resource_id):
    '''Returns the resource group name from parsing the resource id.
    :param str resource_id: The resource id
    '''
    resource_id = resource_id.lower()
    resource_group_keyword = '/resourcegroups/'
    return resource_id[resource_id.index(resource_group_keyword) + len(resource_group_keyword):
                       resource_id.index('/providers/')]

def get_resource_group_name_by_registry_name(registry_name):
    '''Returns the resource group name for the container registry.
    :param str registry_name: The name of container registry
    '''
    arm_resource = _arm_get_resource_by_name(registry_name, ACR_RESOURCE_TYPE)
    return get_resource_group_name_by_resource_id(arm_resource.id)

def get_resource_group_name_by_storage_account_name(storage_account_name):
    '''Returns the resource group name for the storage account.
    :param str storage_account_name: The name of storage account
    '''
    arm_resource = _arm_get_resource_by_name(storage_account_name, STORAGE_RESOURCE_TYPE)
    return get_resource_group_name_by_resource_id(arm_resource.id)

def get_registry_by_name(registry_name, resource_group_name=None):
    '''Returns a tuple of Registry object and resource group name.
    :param str registry_name: The name of container registry
    :param str resource_group_name: The name of resource group
    '''
    if resource_group_name is None:
        resource_group_name = get_resource_group_name_by_registry_name(registry_name)

    client = get_acr_service_client().registries

    return client.get_properties(resource_group_name, registry_name), resource_group_name

def get_access_key_by_storage_account_name(storage_account_name, resource_group_name=None):
    '''Returns access key for the storage account.
    :param str storage_account_name: The name of storage account
    :param str resource_group_name: The name of resource group
    '''
    if resource_group_name is None:
        resource_group_name = get_resource_group_name_by_storage_account_name(storage_account_name)

    client = get_storage_service_client().storage_accounts

    return client.list_keys(resource_group_name, storage_account_name).keys[0].value #pylint: disable=no-member

def arm_deploy_template(resource_group_name,
                        registry_name,
                        location,
                        storage_account_name,
                        admin_user_enabled):
    '''Deploys ARM template to create a container registry with a new storage account.
    :param str resource_group_name: The name of resource group
    :param str registry_name: The name of container registry
    :param str location: The name of location
    :param str storage_account_name: The name of storage account
    :param bool admin_user_enabled: Enable admin user
    '''
    from azure.mgmt.resource.resources.models import DeploymentProperties
    from azure.cli.core._util import get_file_json
    import os

    parameters = _parameters(registry_name, location, storage_account_name, admin_user_enabled)

    file_path = os.path.join(os.path.dirname(__file__), 'template.json')
    template = get_file_json(file_path)
    properties = DeploymentProperties(template=template, parameters=parameters, mode='incremental')

    return _arm_deploy_template(
        get_arm_service_client().deployments, resource_group_name, properties)

def _arm_deploy_template(deployments_client,
                         resource_group_name,
                         properties,
                         index=0):
    '''Deploys ARM template to create a container registry.
    :param obj deployments_client: ARM deployments service client
    :param str resource_group_name: The name of resource group
    :param DeploymentProperties properties: The properties of a deployment
    :param int index: The index added to deployment name to avoid conflict
    '''
    if index == 0:
        deployment_name = ACR_RESOURCE_PROVIDER
    elif index > 9: # Just a number to avoid infinite loops
        raise CLIError(
            'The resource group {} has too many deployments'.format(resource_group_name))
    else:
        deployment_name = ACR_RESOURCE_PROVIDER + '_' + str(index)

    try:
        deployments_client.validate(
            resource_group_name, deployment_name, properties)
        return deployments_client.create_or_update(
            resource_group_name, deployment_name, properties)
    except: #pylint: disable=bare-except
        return _arm_deploy_template(
            deployments_client, resource_group_name, properties, index + 1)

def _parameters(registry_name,
                location,
                storage_account_name,
                admin_user_enabled):
    '''Returns a dict of deployment parameters.
    :param str registry_name: The name of container registry
    :param str location: The name of location
    :param str storage_account_name: The name of storage account
    :param bool admin_user_enabled: Enable admin user
    '''
    parameters = {
        'registryName': {'value': registry_name},
        'registryLocation': {'value': location},
        'storageAccountName': {'value': storage_account_name},
        'adminUserEnabled': {'value': admin_user_enabled},
    }
    customized_api_version = get_acr_api_version()
    if customized_api_version:
        parameters['registryApiVersion'] = {'value': customized_api_version}
    return parameters

def random_storage_account_name(registry_name):
    from datetime import datetime

    client = get_storage_service_client().storage_accounts

    while True:
        time_stamp_suffix = datetime.utcnow().strftime('%H%M%S')
        storage_account_name = ''.join([registry_name[:18], time_stamp_suffix])[:24]
        if client.check_name_availability(storage_account_name).name_available: #pylint: disable=no-member
            return storage_account_name



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from ._factory import get_acr_service_client
from ._utils import get_resource_group_name_by_registry_name

import azure.cli.core._logging as _logging
logger = _logging.get_az_logger(__name__)

def acr_credential_show(registry_name, resource_group_name=None):
    '''Gets the administrator login credentials for the specified container registry.
    :param str registry_name: The name of container registry
    :param str resource_group_name: The name of resource group
    '''
    if resource_group_name is None:
        resource_group_name = get_resource_group_name_by_registry_name(registry_name)

    client = get_acr_service_client().registries

    return client.get_credentials(resource_group_name, registry_name)

def acr_credential_renew(registry_name, resource_group_name=None):
    '''Regenerates the administrator login credentials for the specified container registry.
    :param str registry_name: The name of container registry
    :param str resource_group_name: The name of resource group
    '''
    if resource_group_name is None:
        resource_group_name = get_resource_group_name_by_registry_name(registry_name)

    client = get_acr_service_client().registries

    return client.regenerate_credentials(resource_group_name, registry_name)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import requests

from ._utils import (
    get_registry_by_name
)
from .credential import acr_credential_show

def _obtain_data_from_registry(login_server, path, resultIndex, username, password):
    registryEndpoint = 'https://' + login_server
    resultList = []
    executeNextHttpCall = True

    while executeNextHttpCall:
        executeNextHttpCall = False
        response = requests.get(
            registryEndpoint + path,
            auth=requests.auth.HTTPBasicAuth(
                username,
                password
            )
        )

        if response.status_code == 200:
            resultList += response.json()[resultIndex]
            if 'link' in response.headers and response.headers['link']:
                linkHeader = response.headers['link']
                # The registry is telling us there's more items in the list,
                # and another call is needed. The link header looks something
                # like `Link: </v2/_catalog?last=hello-world&n=1>; rel="next"`
                # we should follow the next path indicated in the link header
                path = linkHeader[(linkHeader.index('<')+1):linkHeader.index('>')]
                executeNextHttpCall = True
        else:
            response.raise_for_status()

    return resultList

def _validate_user_credentials(registry_name, path, resultIndex, username=None, password=None):
    registry, _ = get_registry_by_name(registry_name)
    login_server = registry.login_server #pylint: disable=no-member

    if username:
        if not password:
            import getpass
            password = getpass.getpass('Password: ')
        return _obtain_data_from_registry(login_server, path, resultIndex, username, password)

    try:
        cred = acr_credential_show(registry_name)
        username = cred.username
        password = cred.password
        return _obtain_data_from_registry(login_server, path, resultIndex, username, password)
    except: #pylint: disable=bare-except
        pass

    try:
        user_input = raw_input
    except NameError:
        user_input = input

    import getpass
    username = user_input("Username: ")
    password = getpass.getpass('Password: ')
    return _obtain_data_from_registry(login_server, path, resultIndex, username, password)

def acr_repository_list(registry_name, username=None, password=None):
    '''Lists repositories in the specified container registry.
    :param str registry_name: The name of container registry
    :param str username: The username used to log into the container registry
    :param str password: The password used to log into the container registry
    '''
    path = '/v2/_catalog'
    return _validate_user_credentials(registry_name, path, 'repositories', username, password)

def acr_repository_show_tags(registry_name, repository, username=None, password=None):
    '''Shows tags of a given repository in the specified container registry.
    :param str registry_name: The name of container registry
    :param str repository: The repository to obtain tags from
    :param str username: The username used to log into the container registry
    :param str password: The password used to log into the container registry
    '''
    path = '/v2/' + repository + '/tags/list'
    return _validate_user_credentials(registry_name, path, 'tags', username, password)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.core.commands import register_cli_argument
from azure.cli.core.commands.parameters import (
    resource_group_name_type,
    location_type,
    tags_type,
    get_resource_name_completion_list
)

from ._constants import (
    ACR_RESOURCE_TYPE,
    STORAGE_RESOURCE_TYPE
)
from ._validators import (
    validate_resource_group_name,
    validate_registry_name
)

register_cli_argument('acr', 'registry_name',
                      options_list=('--name', '-n'),
                      help='The name of the container registry',
                      completer=get_resource_name_completion_list(ACR_RESOURCE_TYPE))
register_cli_argument('acr', 'storage_account_name',
                      help='The name of an existing storage account',
                      completer=get_resource_name_completion_list(STORAGE_RESOURCE_TYPE))

register_cli_argument('acr', 'resource_group_name', resource_group_name_type)
register_cli_argument('acr', 'location', location_type)
register_cli_argument('acr', 'tags', tags_type)
register_cli_argument('acr', 'admin_enabled',
                      help='Indicates whether the admin user is enabled',
                      choices=['true', 'false'])

register_cli_argument('acr', 'username',
                      options_list=('--username', '-u'),
                      help='The username used to log into a container registry')
register_cli_argument('acr', 'password',
                      options_list=('--password', '-p'),
                      help='The password used to log into a container registry')

register_cli_argument('acr create', 'registry_name', completer=None,
                      validator=validate_registry_name)
register_cli_argument('acr create', 'resource_group_name',
                      validator=validate_resource_group_name)
register_cli_argument('acr check-name', 'registry_name', completer=None)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.core.commands import LongRunningOperation

from azure.mgmt.containerregistry.models import (
    Registry,
    RegistryUpdateParameters,
    StorageAccountProperties
)

from ._factory import get_acr_service_client
from ._utils import (
    get_access_key_by_storage_account_name,
    get_resource_group_name_by_registry_name,
    arm_deploy_template,
    random_storage_account_name
)

import azure.cli.core._logging as _logging
logger = _logging.get_az_logger(__name__)

def acr_check_name(registry_name):
    '''Checks whether the container registry name is available for use.
    :param str registry_name: The name of container registry
    '''
    client = get_acr_service_client().registries

    return client.check_name_availability(registry_name)

def acr_list(resource_group_name=None):
    '''Lists all the available container registries under the current subscription.
    :param str resource_group_name: The name of resource group
    '''
    client = get_acr_service_client().registries

    if resource_group_name:
        return client.list_by_resource_group(resource_group_name)
    else:
        return client.list()

def acr_create(registry_name, #pylint: disable=too-many-arguments
               resource_group_name,
               location,
               storage_account_name=None,
               admin_enabled='false'):
    '''Creates a container registry.
    :param str registry_name: The name of container registry
    :param str resource_group_name: The name of resource group
    :param str location: The name of location
    :param str storage_account_name: The name of storage account
    :param str admin_enabled: Indicates whether the admin user is enabled
    '''
    client = get_acr_service_client().registries
    admin_user_enabled = admin_enabled == 'true'

    if storage_account_name is None:
        storage_account_name = random_storage_account_name(registry_name)
        LongRunningOperation()(
            arm_deploy_template(resource_group_name,
                                registry_name,
                                location,
                                storage_account_name,
                                admin_user_enabled)
        )
        registry = client.get_properties(resource_group_name, registry_name)
    else:
        storage_account_key = get_access_key_by_storage_account_name(storage_account_name)
        registry = client.create_or_update(
            resource_group_name, registry_name,
            Registry(
                location=location,
                storage_account=StorageAccountProperties(
                    storage_account_name,
                    storage_account_key
                ),
                admin_user_enabled=admin_user_enabled
            )
        )

    logger.warning('\nCreate a new service principal and assign access:')
    logger.warning(
        '  az ad sp create-for-rbac --scopes %s --role Owner --password <password>',
        registry.id) #pylint: disable=no-member
    logger.warning('\nUse an existing service principal and assign access:')
    logger.warning(
        '  az role assignment create --scope %s --role Owner --assignee <app-id>',
        registry.id) #pylint: disable=no-member

    return registry

def acr_delete(registry_name, resource_group_name=None):
    '''Deletes a container registry.
    :param str registry_name: The name of container registry
    :param str resource_group_name: The name of resource group
    '''
    if resource_group_name is None:
        resource_group_name = get_resource_group_name_by_registry_name(registry_name)

    client = get_acr_service_client().registries

    return client.delete(resource_group_name, registry_name)

def acr_show(registry_name, resource_group_name=None):
    '''Gets the properties of the specified container registry.
    :param str registry_name: The name of container registry
    :param str resource_group_name: The name of resource group
    '''
    if resource_group_name is None:
        resource_group_name = get_resource_group_name_by_registry_name(registry_name)

    client = get_acr_service_client().registries

    return client.get_properties(resource_group_name, registry_name)

def acr_update_get(client,
                   registry_name,
                   resource_group_name=None):
    if resource_group_name is None:
        resource_group_name = get_resource_group_name_by_registry_name(registry_name)

    props = client.get_properties(resource_group_name, registry_name)

    return RegistryUpdateParameters(
        tags=props.tags,
        admin_user_enabled=props.admin_user_enabled,
        storage_account=props.storage_account
    )

def acr_update_custom(instance,
                      admin_enabled=None,
                      storage_account_name=None,
                      tags=None):
    if admin_enabled is not None:
        instance.admin_user_enabled = admin_enabled == 'true'

    if tags is not None:
        instance.tags = tags

    if storage_account_name is not None:
        storage_account_key = \
            get_access_key_by_storage_account_name(storage_account_name)
        storage_account = StorageAccountProperties(
            storage_account_name,
            storage_account_key
        )
    instance.storage_account = storage_account if storage_account_name else None

    return instance

def acr_update_set(client,
                   registry_name,
                   resource_group_name=None,
                   parameters=None):
    if resource_group_name is None:
        resource_group_name = get_resource_group_name_by_registry_name(registry_name)

    return client.update(resource_group_name, registry_name, parameters)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

#pylint: disable=line-too-long

from azure.cli.core.commands import cli_command
from azure.cli.core.commands.arm import cli_generic_update_command

from ._format import output_format
from ._factory import get_acr_service_client

cli_command(__name__, 'acr credential show', 'azure.cli.command_modules.acr.credential#acr_credential_show', table_transformer=output_format)
cli_command(__name__, 'acr credential renew', 'azure.cli.command_modules.acr.credential#acr_credential_renew', table_transformer=output_format)

cli_command(__name__, 'acr check-name', 'azure.cli.command_modules.acr.custom#acr_check_name')
cli_command(__name__, 'acr list', 'azure.cli.command_modules.acr.custom#acr_list', table_transformer=output_format)
cli_command(__name__, 'acr create', 'azure.cli.command_modules.acr.custom#acr_create', table_transformer=output_format)
cli_command(__name__, 'acr delete', 'azure.cli.command_modules.acr.custom#acr_delete', table_transformer=output_format)
cli_command(__name__, 'acr show', 'azure.cli.command_modules.acr.custom#acr_show', table_transformer=output_format)
cli_generic_update_command(__name__,
                           'acr update',
                           'azure.cli.command_modules.acr.custom#acr_update_get',
                           'azure.cli.command_modules.acr.custom#acr_update_set',
                           factory=lambda: get_acr_service_client().registries,
                           custom_function_op='azure.cli.command_modules.acr.custom#acr_update_custom',
                           table_transformer=output_format)

cli_command(__name__, 'acr repository list', 'azure.cli.command_modules.acr.repository#acr_repository_list')
cli_command(__name__, 'acr repository show-tags', 'azure.cli.command_modules.acr.repository#acr_repository_show_tags')
